def is_pending(hub, ret: dict):
    # The azure resource contains provisioningState. the optional values are:
    # CANCELED, CREATING, DELETING, FAILED, SUCCEEDED, UNKNOWN and UPDATING
    # provisioningState 'Succeeded' can be return when a new absent or present/update
    # request has not yet started. Therefore require 'Succeeded' without changes.
    hub.log.debug("Azure is_pending()")

    new_state = ret.get("new_state", None)
    status = None
    if new_state:
        status = _search_value(new_state, "provisioningState")

    if status and isinstance(status, str):
        hub.log.debug(f"Azure is_pending() status {status}")
        if status.casefold() == "failed" or status.casefold() == "canceled":
            hub.log.debug(
                f"No need to reconcile new state {new_state} with status {status}"
            )
            return False
        if (
            status.casefold() == "creating"
            or status.casefold() == "updating"
            or status.casefold() == "deleting"
        ):
            hub.log.debug(
                f"Need to reconcile new state {new_state} with status {status}"
            )
            return True

    # Reconcile for 'False' result or if there are changes
    return not ret["result"] is True or bool(ret["changes"])


def _search_value(obj, key):
    """Return a value corresponding to the specified key in the (possibly
    nested) dictionary d. If there is no item with that key, return
    default.
    """
    if key in obj:
        return obj[key]
    for k, v in obj.items():
        if isinstance(v, dict):
            item = _search_value(v, key)
            if item is not None:
                return item



from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Get a new access token based on Azure client_id, client_secret and tenant_id

    Example:
    .. code-block:: yaml

        azure:
          profile_name:
            client_id: Azure client id "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
            secret: Azure client secret "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
            subscription_id: Azure subscription id "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
            tenant: Azure tenant id "cccccccc-cccc-cccc-cccc-cccccccccccc"
            api_version: 2015-07-01
    """
    sub_profiles = {}
    for (
        profile,
        ctx,
    ) in profiles.get("azure", {}).items():
        temp_ctx = NamespaceDict(acct={})
        tenant = ctx["tenant"]
        client_id = ctx["client_id"]
        secret = ctx["secret"]
        subscription_id = ctx["subscription_id"]

        ret = await hub.exec.request.json.post(
            temp_ctx,
            url=f"https://login.microsoftonline.com/{tenant}/oauth2/token",
            headers={
                "content-type": "application/x-www-form-urlencoded",
            },
            data={
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": secret,
                "resource": "https://management.azure.com/",
            },
        )

        if not ret["status"]:
            comment = ret.get("comment", "")
            error = f"Unable to authenticate with tenant '{tenant}' and client id '{client_id}': {comment}"
            hub.log.error(error)
            raise ConnectionError(error)

        access_token = ret["ret"]["access_token"]
        sub_profiles[profile] = dict(
            subscription_id=subscription_id,
            endpoint_url=hub.exec.azure.URL,
            headers={"Authorization": f"Bearer {access_token}"},
        )
    return sub_profiles



from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


async def gather(hub, profile: Dict[str, Any]):
    identity_ctx = NamespaceDict(acct=profile)
    return {
        "attribute_value": identity_ctx["acct"]["subscription_id"],
        "attribute_key": "subscription_id",
        "provider": "AZURE",
    }
    return {}



import copy
import json
from typing import Any
from typing import Dict


def convert_raw_management_group_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    management_group_name: str,
    resource_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        management_group_name: Azure Resource Group name.
        resource_id: Azure Managment group resource id.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "management_group_name": management_group_name,
        "resource_id": resource_id,
    }
    if type(resource) == str:
        resource = json.loads(resource)
    properties = resource.get("properties")
    if properties:
        if "displayName" in properties:
            resource_translated["display_name"] = properties["displayName"]
        if "details" in properties:
            resource_translated["parent_id"] = properties["details"]["parent"]["id"]

    return resource_translated


def convert_present_to_raw_management_group(hub, display_name: str, parent_id: str):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        display_name: The management group name to be displayed
        parent_id: creates management group under this id

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {}
    if display_name is not None or parent_id is not None:
        payload["properties"] = {}
    if display_name is not None:
        payload["properties"]["displayName"] = display_name
    if parent_id is not None:
        payload["properties"]["details"] = {"parent": {"id": parent_id}}

    return payload


def update_management_groups_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (
        (new_values.get("parent_id") is not None)
        and (existing_payload.get("properties") is not None)
        and (existing_payload.get("properties").get("details") is not None)
        and (
            existing_payload.get("properties").get("details").get("parent") is not None
        )
        and (
            existing_payload.get("properties").get("details").get("parent").get("id")
            != new_values.get("parent_id")
        )
    ):
        new_payload["properties"]["details"]["parent"]["id"] = new_values["parent_id"]
        need_update = True
    if (new_values.get("display_name") is not None) and (
        existing_payload.get("properties").get("displayName")
        != new_values.get("display_name")
    ):
        new_payload["properties"]["displayName"] = new_values["display_name"]
        need_update = True

    if need_update:
        result["ret"] = new_payload
    return result



import re


try:
    import bs4
    import requests

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def parse(hub, url: str):
    """
    Parse a page such as this one:
    https://docs.microsoft.com/en-us/rest/api/network-gateway/local-network-gateways/create-or-update
    Return a dictionary of a REST api spec"
    """
    plugin = {}
    with requests.get(url) as response:
        html = response.text

    soup = bs4.BeautifulSoup(html, "lxml")

    dts = soup.find_all("dt")
    for dt in dts:
        # Fetch the name of the service that the api belongs to
        if re.sub(r"[\t\n]", "", dt.text) == "Service:":
            plugin.update(
                {"service": re.sub(r"[\t\n]", "", dt.find_next_sibling("dd").text)}
            )
        # Fetch the api version
        if re.sub(r"[\t\n]", "", dt.text) == "API Version:":
            plugin.update(
                {"api_version": re.sub(r"[\t\n]", "", dt.find_next_sibling("dd").text)}
            )

    http_spec = soup.find(class_="lang-http").text
    if http_spec is not None:
        api_method = http_spec.split()[0]
        hub.tool.validators.http.request_method(api_method)
        plugin.update({"api_method": api_method})
        api_url = http_spec.split()[1]
        hub.tool.validators.http.url(api_url)
        plugin.update({"api_url": api_url})
    else:
        err_msg = f"Cannot find api's http specs at {url}"
        hub.log.error("Error: " + err_msg)
        raise RuntimeError(err_msg)

    # Fetch uri parameters of the api url
    uri_param_header = soup.find(id="uri-parameters")
    if uri_param_header is not None:
        uri_parameters = parse_uri_parameters(
            uri_param_header.find_next_sibling("table")
        )
        if uri_parameters:
            plugin.update({"uri_parameters": uri_parameters})
            # Update api url to snake case
            for uri_parameter in uri_parameters:
                api_url = api_url.replace(
                    uri_parameter.get("name"), uri_parameter.get("name_formatted")
                )
            plugin.update({"api_url": api_url})

    # Fetch the request body of the api
    request_body_header = soup.find(id="request-body")
    if request_body_header is not None:
        request_body = parse_request_body(
            request_body_header.find_next_sibling("table")
        )
        if request_body:
            plugin.update({"request_body": request_body})

    # Fetch the response of the api
    responses_header = soup.find(id="response")
    if responses_header is not None:
        responses = parse_responses(responses_header.find_next_sibling("table"))
        if responses:
            plugin.update({"responses": responses})

    # Fetch definitions of object types used in the api
    definitions_header = soup.find(id="definitions")
    definitions = {}
    if definitions_header is not None:
        definitions_name = parse_definitions(
            definitions_header.find_next_sibling("table")
        )
        definition_header_list = soup.find_all("h3")
        for definition_name in definitions_name:
            definition_header = next(
                filter(
                    lambda h3: definition_name == re.sub(r"[\t\n]", "", h3.text),
                    definition_header_list,
                ),
                None,
            )
            if definition_header is not None:
                definition = parse_definition(
                    definition_header.find_next_sibling("table")
                )
                definitions.update({definition_name: definition})
    if definitions:
        plugin.update({"definitions": definitions})

    return plugin


def parse_uri_parameters(node):
    parse_results = []
    rows = node.find_all("tr")
    for row in rows:
        if row.find("th") is not None:
            continue
        cols = row.find_all("td")
        parameter_name = re.sub(r"[\t\n]", "", cols[0].text)
        parameter_in = re.sub(r"[\t\n]", "", cols[1].text)
        parameter_required = re.sub(r"[\t\n]", "", cols[2].text)
        parameter_type = re.sub(r"[\t\n]", "", cols[3].text)
        parameter_description = re.sub(r"[\t\n]", "", cols[4].text)
        parse_results.append(
            {
                "name": parameter_name,
                "name_formatted": re.sub(
                    "(?<!^)(?=[A-Z])", "_", parameter_name
                ).lower(),
                "in": parameter_in,
                "required": parameter_required,
                "type": parameter_type,
                "description": parameter_description,
            }
        )
    return parse_results


def parse_request_body(node):
    parse_result = []
    rows = node.find_all("tr")
    for row in rows:
        if row.find("th") is not None:
            continue
        cols = row.find_all("td")
        request_body_name = re.sub(r"[\t\n]", "", cols[0].text)
        request_body_type = re.sub(r"[\t\n]", "", cols[1].text)
        request_body_description = re.sub(r"[\t\n]", "", cols[2].text)
        parse_result.append(
            {
                "name": request_body_name,
                "type": request_body_type,
                "description": request_body_description,
            }
        )
    return parse_result


def parse_responses(node):
    parse_results = []
    rows = node.find_all("tr")
    for row in rows:
        if row.find("th") is not None:
            continue
        cols = row.find_all("td")
        response_name = re.sub(r"[\t\n]", "", cols[0].text)
        http_status_code = None
        if response_name.split()[0].isnumeric():
            http_status_code = int(response_name.split()[0])
            http_status_info = response_name.split()[1]
        else:
            http_status_info = response_name
        response_type = re.sub(r"[\t\n]", "", cols[1].text)
        response_description = re.sub(r"[\t\n]", "", cols[2].text)
        parse_results.append(
            {
                "http_status_code": http_status_code,
                "http_status_info": http_status_info,
                "type": response_type,
                "description": response_description,
            }
        )
    return parse_results


def parse_definitions(node):
    parse_results = []
    rows = node.find_all("tr")
    for row in rows:
        cols = row.find_all("td")
        definition = re.sub(r"[\t\n]", "", cols[0].text)
        parse_results.append(definition)
    return parse_results


def parse_definition(node):
    parse_results = []
    rows = node.find_all("tr")
    for row in rows:
        if row.find("th") is not None:
            continue
        cols = row.find_all("td")
        definition_name = re.sub(r"[\t\n]", "", cols[0].text)
        definition_type = re.sub(r"[\t\n]", "", cols[1].text)
        definition_description = re.sub(r"[\t\n]", "", cols[2].text)
        parse_results.append(
            {
                "name": definition_name,
                "type": definition_type,
                "description": definition_description,
            }
        )
    return parse_results



import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_security_group_name: str,
    location: str,
    security_rules: List[
        make_dataclass(
            "SecurityRule",
            [
                ("name", str),
                ("priority", int),
                ("direction", str),
                ("access", str),
                ("protocol", str),
                ("source_port_range", str, field(default=None)),
                ("destination_port_range", str, field(default=None)),
                ("source_address_prefix", str, field(default=None)),
                ("destination_address_prefix", str, field(default=None)),
                ("source_port_ranges", List, field(default=None)),
                ("destination_port_ranges", List, field(default=None)),
                ("source_address_prefixes", List, field(default=None)),
                ("destination_address_prefixes", List, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""
    Create or update Network Security Groups

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_security_group_name(str): The name of the network security group.
        location(str): The location to create the resource in.
        security_rules(List, optional): List of security rules.Either source_port_range or source_port_ranges needs to be provided.
                                        Either destination_port_range or destination_port_ranges needs to be provided.
                                        Either source_address_prefix or source_address_prefixes needs to be provided.
                                        Either destination_address_prefix or destination_address_prefixs needs to be provided
        tags(Dict, optional): Resource tags.
        subscription_id(str, optional): Subscription Unique id.
        resource_id(str, optional): Resource Group id on Azure.

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            resource_is_present:
              azure.network.network_security_groups.present:
                - name: value
                - resource_group_name: value
                - network_security_group_name: value
                - location: value
                - tags: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}"

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{hub.exec.azure.URL}{resource_id}?api-version=2022-07-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] == 404:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "network_security_group_name": network_security_group_name,
                        "subscription_id": subscription_id,
                        "security_rules": security_rules,
                        "tags": tags,
                        "location": location,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.network.network_security_groups '{name}'"
                )
                return result

            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.network.network_security_groups.convert_present_to_raw_network_security_groups(
                    location=location,
                    security_rules=security_rules,
                    tags=tags,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{hub.exec.azure.URL}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=payload,
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create Network Security Groups {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result
                result[
                    "new_state"
                ] = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                    network_security_group_name=network_security_group_name,
                    resource_group_name=resource_group_name,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Created azure.network.network_security_groups '{name}'"
                )
                return result

        else:
            hub.log.debug(
                f"Could not get azure.network.network_security_groups {response_get['comment']} {response_get['ret']}"
            )
            result["comment"] = [response_get["comment"], response_get["ret"]]
            result["result"] = False
            return result

    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            resource_id=resource_id,
            network_security_group_name=network_security_group_name,
            resource_group_name=resource_group_name,
            subscription_id=subscription_id,
        )
        # Generate a new PUT operation payload with new values to update
        new_payload = hub.tool.azure.network.network_security_groups.update_network_security_groups_payload(
            existing_resource,
            {"tags": tags, "security_rules": security_rules},
        )
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.network.network_security_groups '{name}' has no property need to be updated."
            )
            return result
        else:
            if ctx.get("test", False):
                result[
                    "new_state"
                ] = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                    network_security_group_name=network_security_group_name,
                    resource_group_name=resource_group_name,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Would update azure.network.network_security_groups '{name}'"
                )
                return result
            else:
                # PUT call to update the resource
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{hub.exec.azure.URL}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=new_payload["ret"],
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not update azure.network.network_security_groups {response_put['comment']} {response_put['ret']}"
                    )
                    result["result"] = False
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                    network_security_group_name=network_security_group_name,
                    resource_group_name=resource_group_name,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Updated azure.network.network_security_groups '{name}'"
                )
                return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    network_security_group_name: str,
    subscription_id: str = None,
) -> Dict:
    r"""
    Delete Network Security Groups

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        network_security_group_name(str): The name of the network security group.
        subscription_id: Unique subscription ID

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.network.network_security_groups.absent:
                - name: value
                - resource_group_name: value
                - network_security_group_name: value
    """

    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id

    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}"

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{hub.exec.azure.URL}{resource_id}?api-version=2022-07-01",
        success_codes=[200],
    )
    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            resource_id=resource_id,
            network_security_group_name=network_security_group_name,
            resource_group_name=resource_group_name,
            subscription_id=subscription_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.network.network_security_groups '{name}'"
            )
            return result

        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{hub.exec.azure.URL}{resource_id}?api-version=2022-07-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.network.network_security_groups {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(
            f"Deleted azure.network.network_security_groups '{name}'"
        )
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.network.network_security_groups '{name}' already absent"
        )
        return result
    else:
        hub.log.debug(
            f"Could not get azure.network.network_security_groups {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    List all Network Security Groups under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.network.network_security_groups
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "networkSecurityGroups": "network_security_group_name",
            "subscriptions": "subscription_id",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{hub.exec.azure.URL}/subscriptions/{subscription_id}/providers/Microsoft.Network/networkSecurityGroups?api-version=2022-07-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource["id"], uri_parameters
                )
                translated_resource = hub.tool.azure.network.network_security_groups.convert_raw_network_security_groups_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    **uri_parameter_values,
                )
                result[resource_id] = {
                    f"azure.network.network_security_groups.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    virtual_network_name: str,
    subnet_name: str,
    address_prefix: str,
    enforce_private_link_endpoint_network_policies: bool = None,
    delegations: List[
        make_dataclass(
            "DelegationSet",
            [
                ("name", str, field(default=None)),
                ("service", str, field(default=None)),
            ],
        )
    ] = None,
    enforce_private_link_service_network_policies: bool = None,
    service_endpoints: List = None,
    service_endpoint_policy_ids: List = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    rf"""
    **Autogenerated function**

    Create Subnets

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        subnet_name(str): The name of the subnet.
        address_prefix(str): The address prefix for the subnet.
        enforce_private_link_endpoint_network_policies(bool, optional): Enable or Disable apply network policies on private end point in the subnet.
        delegations(List, optional): An array of references to the delegations on the subnet.
         Defaults to None
         * name(str) -- The name of the resource that is unique within a subnet.
         * service(str) -- The name of the service to whom the subnet should be delegated.
        enforce_private_link_service_network_policies(bool, optional): Enable or Disable apply network policies on private link service in the subnet.
        service_endpoints(List, optional): An array of service endpoints
        service_endpoint_policy_ids(List, optional): An array of service endpoint policy Ids.\
        subscription_id(str,optional): Subscription Unique id.
        resource_id(str, optional): Subnet resource id on Azure


    Returns:
        Dict

    Examples:

        .. code-block:: sls

            my-subnet:
              azure.network.subnets.present:
                - name: my-subnet
                - resource_group_name: my-resource-group
                - subscription_id: my-subscription
                - virtual_network_name: my-vnet
                - subnet_name: my-subnet-1
                - address_prefix: 10.0.0.0/24
                - enforce_private_link_endpoint_network_policies: True/False
                - delegations:
                    - name: my-delegation-1
                      service: Microsoft.StoragePool/diskPools
                - enforce_private_link_service_network_policies: True/False
                - service_endpoints:
                    - Microsoft.Sql
                    - Microsoft.Storage
                - service_endpoint_policy_ids:
                    - /subscriptions/subscription_id/resourceGroups/resource_group_name/providers/Microsoft.Network/serviceEndpointPolicies/service_endpoint_policy_name
                    - /subscriptions/subscription_id/resourceGroups/resource_group_name/providers/Microsoft.Network/serviceEndpointPolicies/service_endpoint_policy_name

    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}/subnets/{subnet_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] == 404:
            # Return a proposed state by Idem state --test
            if ctx.get("test", False):
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "subscription_id": subscription_id,
                        "virtual_network_name": virtual_network_name,
                        "subnet_name": subnet_name,
                        "address_prefix": address_prefix,
                        "enforce_private_link_endpoint_network_policies": enforce_private_link_endpoint_network_policies,
                        "delegations": delegations,
                        "enforce_private_link_service_network_policies": enforce_private_link_service_network_policies,
                        "service_endpoints": service_endpoints,
                        "service_endpoint_policy_ids": service_endpoint_policy_ids,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(f"Would create azure.network.subnets '{name}'")
                return result

            else:
                # PUT operation to create a resource.
                payload = hub.tool.azure.network.subnets.convert_present_to_raw_subnets(
                    address_prefix=address_prefix,
                    enforce_private_link_endpoint_network_policies=enforce_private_link_endpoint_network_policies,
                    delegations=delegations,
                    enforce_private_link_service_network_policies=enforce_private_link_service_network_policies,
                    service_endpoints=service_endpoints,
                    service_endpoint_policy_ids=service_endpoint_policy_ids,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}/subnets/{subnet_name}?api-version=2021-03-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.network.subnets '{name}' {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    virtual_network_name=virtual_network_name,
                    subnet_name=subnet_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(f"Created azure.network.subnets '{name}'")
                return result
        else:
            hub.log.debug(
                f"Could not get azure.network.subnets {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result
    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            subnet_name=subnet_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        # Generate a new PUT operation payload with new values
        new_payload = hub.tool.azure.network.subnets.update_subnets_payload(
            existing_resource,
            {
                "address_prefix": address_prefix,
                "enforce_private_link_endpoint_network_policies": enforce_private_link_endpoint_network_policies,
                "delegations": delegations,
                "enforce_private_link_service_network_policies": enforce_private_link_service_network_policies,
                "service_endpoints": service_endpoints,
                "service_endpoint_policy_ids": service_endpoint_policy_ids,
            },
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.subnets '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    virtual_network_name=virtual_network_name,
                    subnet_name=subnet_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(f"Would update azure.network.subnets '{name}'")
            return result

        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.network.subnets '{name}' has no property need to be updated."
            )
            return result
        result["comment"] = result["comment"] + new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 201],
            json=new_payload["ret"],
        )
        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.network.subnets {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_put["comment"], response_put["ret"]]
            return result

        result[
            "new_state"
        ] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            subnet_name=subnet_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        result["comment"].append(f"Updated azure.network.subnets '{name}'")
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    virtual_network_name: str,
    subnet_name: str,
    subscription_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Delete Subnets

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        subnet_name(str): The name of the subnet.
        subscription_id(str,optional): Subscription Unique id.

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            my-subnet:
              azure.network.subnets.absent:
                - name: my-subnet
                - resource_group_name: my-resource-group
                - virtual_network_name: my-vnet
                - subnet_name: my-subnet-1
                - subscription_id: my-subscription
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}/subnets/{subnet_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )
    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            subnet_name=subnet_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        if ctx.get("test", False):
            result["comment"].append(f"Would delete azure.network.subnets '{name}'")
            return result

        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.network.subnets{response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.network.subnets '{name}'")
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(f"azure.network.subnets '{name}' already absent")
        return result
    else:
        hub.log.debug(
            f"Could not get azure.network.subnets '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Subnets under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.network.subnets
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "virtualNetworks": "virtual_network_name",
            "subnets": "subnet_name",
            "subscriptions": "subscription_id",
        }
    )
    async for vnet_page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/virtualNetworks?api-version=2021-03-01",
        success_codes=[200],
    ):
        vnet_list = vnet_page_result.get("value")
        if vnet_list:
            for vnet in vnet_list:
                if vnet.get("properties") and vnet.get("properties").get("subnets"):
                    subnet_list = vnet["properties"]["subnets"]
                    for resource in subnet_list:
                        uri_parameter_values = (
                            hub.tool.azure.uri.get_parameter_value_in_dict(
                                resource["id"], uri_parameters
                            )
                        )
                        resource_translated = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
                            resource=resource,
                            idem_resource_name=resource["id"],
                            resource_id=resource["id"],
                            **uri_parameter_values,
                        )
                        result[resource["id"]] = {
                            f"azure.network.subnets.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in resource_translated.items()
                            ]
                        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    virtual_network_name: str,
    address_space: List,
    location: str,
    resource_id: str = None,
    bgp_communities: Dict = None,
    flow_timeout_in_minutes: int = None,
    subnets: List[
        make_dataclass(
            "SubnetSet",
            [
                ("name", str, field(default=None)),
                ("address_prefix", str, field(default=None)),
                ("security_group_id", str, field(default=None)),
                ("service_endpoints", List, field(default=None)),
            ],
        )
    ] = None,
    subscription_id: str = None,
    tags: Dict = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Create or update Virtual Networks

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        address_space(list): An array of IP address ranges that can be used by subnets of the virtual network.
        location(str): Resource location. This field can not be updated.
        resource_id(str, optional): Virtual Network resource id on Azure
        bgp_communities(dict, optional): Bgp Communities sent over ExpressRoute with each route corresponding to a prefix in this VNET.
        flow_timeout_in_minutes(int, optional): The FlowTimeout value (in minutes) for the Virtual Network
        subnets(List, optional): List of Subnet in a virtual network resource.Each Subnet will have fields
          name(str, required), address_space(str, required) , security_group_id(str, optional) and service_endpoints(list, optional)
        subscription_id(str,optional): Subscription Unique id.
        tags(dict, optional): Resource tags.

    Returns:
        dict

    Examples:

        .. code-block:: sls

            my-vnet:
              azure.network.virtual_networks.present:
                - name: my-vnet
                - resource_group_name: my-rg-1
                - virtual_network_name: my-vnet-1
                - location: westus
                - flow_timeout_in_minutes: 15
                - tags:
                    my-tag-key: my-tag-value
                - subnets:
                    - name: subnet_name
                      address_prefix: 10.0.0.0/26
                      security_group_id: /subscriptions/subscription_id/resourceGroups/resource_group_name/providers/Microsoft.Network/networkSecurityGroups/network-security-group-name
                      service_endpoints:
                          - Microsoft.Storage
                - address_space:
                    - 10.0.0.0/26
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )

    if not response_get["result"]:
        if response_get["status"] == 404:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "virtual_network_name": virtual_network_name,
                        "address_space": address_space,
                        "tags": tags,
                        "location": location,
                        "resource_id": resource_id,
                        "flow_timeout_in_minutes": flow_timeout_in_minutes,
                        "bgp_communities": bgp_communities,
                        "subnets": subnets,
                        "subscription_id": subscription_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.network.virtual_networks '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.network.virtual_networks.convert_present_to_raw_virtual_network(
                    subscription_id=subscription_id,
                    address_space=address_space,
                    location=location,
                    bgp_communities=bgp_communities,
                    flow_timeout_in_minutes=flow_timeout_in_minutes,
                    subnets=subnets,
                    tags=tags,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
                    success_codes=[200, 201],
                    json=payload,
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.network.virtual_networks {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    virtual_network_name=virtual_network_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Created azure.network.virtual_networks '{name}'"
                )
                return result

        else:
            hub.log.debug(
                f"Could not get azure.network.virtual_networks {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result
    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        # Generate a new PUT operation payload with new values
        new_payload = (
            hub.tool.azure.network.virtual_networks.update_virtual_network_payload(
                subscription_id,
                response_get["ret"],
                {
                    "address_space": address_space,
                    "bgp_communities": bgp_communities,
                    "flow_timeout_in_minutes": flow_timeout_in_minutes,
                    "subnets": subnets,
                    "tags": tags,
                },
            )
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.virtual_networks '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    virtual_network_name=virtual_network_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Would update azure.network.virtual_networks '{name}'"
                )
            return result
        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.network.virtual_networks '{name}' has no property need to be updated."
            )
            return result
        result["comment"] += new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
            success_codes=[200],
            json=new_payload["ret"],
        )

        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.network.virtual_networks {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_put["comment"], response_put["ret"]]
            return result

        result[
            "new_state"
        ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        result["comment"].append(f"Updated azure.network.virtual_networks '{name}'")
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    virtual_network_name: str,
    subscription_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Delete Virtual Networks

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        subscription_id(str,optional): Subscription Unique id.

    Returns:
        dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.network.virtual_networks.absent:
                - name: value
                - resource_group_name: value
                - virtual_network_name: value
    """
    result = dict(name=name, result=True, comment=[], old_state=None, new_state=None)
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )
    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.network.virtual_networks '{name}'"
            )
            return result
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.network.virtual_networks '{name}' {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.network.virtual_networks '{name}'")
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.network.virtual_networks '{name}' already absent"
        )
        return result
    else:
        hub.log.debug(
            f"Could not get azure.network.virtual_networks '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Virtual Networks under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.network.virtual_networks
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "virtualNetworks": "virtual_network_name",
            "subscriptions": "subscription_id",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/virtualNetworks?api-version=2021-03-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                resource_translated = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    **uri_parameter_values,
                )
                result[resource_id] = {
                    f"azure.network.virtual_networks.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }
    return result



from typing import Any
from typing import Dict


def convert_raw_resource_group_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        resource_id: Azure Resource Group id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "location": resource["location"],
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    return resource_translated


def convert_present_to_raw_resource_group(
    hub,
    location: str,
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location: Resource location.
        tags: Resource tags.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {"location": location}
    if tags is not None:
        payload["tags"] = tags
    return payload



from collections import OrderedDict


async def get(hub, ctx, name: str, resource_id: str):
    result = dict(comment=[], ret=None, result=True)
    if resource_id is None or name is None:
        result["result"] = False
        result["comment"].append(
            "name and resource_id cannot be None for azure.authorization.role_assignments get()"
        )
        return result
    before = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/{resource_id}?api-version=2015-07-01",
        success_codes=[200],
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        if before.get("status") == 404:
            return result
        result["result"] = False
        return result
    uri_parameters = OrderedDict({"roleAssignments": "role_assignment_name"})
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )
    result[
        "ret"
    ] = hub.tool.azure.authorization.role_assignments.convert_raw_role_assignments_to_present(
        resource=before["ret"],
        idem_resource_name=name,
        role_assignment_name=uri_parameter_values["role_assignment_name"],
        resource_id=resource_id,
    )
    return result



from typing import Any
from typing import Dict


def convert_raw_role_assignments_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    role_assignment_name: str,
    resource_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        role_assignment_name: Azure role assignment resource name.
        resource_id: Role assignment resource id on Azure.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "role_assignment_name": role_assignment_name,
    }
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "roleDefinitionId": "role_definition_id",
            "principalId": "principal_id",
            "scope": "scope",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties[parameter_raw]
    return resource_translated


def convert_present_to_raw_role_assignments(
    hub,
    role_definition_id: str = None,
    principal_id: int = None,
) -> Dict[str, Any]:
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        role_definition_id: The role definition ID used in the role assignment.
        principal_id: The principal ID assigned to the role. This maps to the ID inside the Active Directory. It can point to a user, service principal, or security group.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {
            "roleDefinitionId": role_definition_id,
            "principalId": principal_id,
        }
    }
    return payload



import copy
from typing import Any
from typing import Dict
from typing import List


def update_network_interfaces_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        new_values.get("tags") != existing_payload.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        need_update = True
    existing_properties = existing_payload["properties"]

    if new_values.get("ip_configurations") is not None:
        existing_ip_configurations = existing_properties.get("ipConfigurations")
        existing_ip_configurations_payload = convert_raw_ip_configurations_to_present(
            ip_configurations=existing_ip_configurations
        )

        if diff_network_ip_configurations(
            new_values.get("ip_configurations"),
            existing_ip_configurations_payload,
            "name",
        ):
            new_payload["properties"][
                "ipConfigurations"
            ] = convert_present_ip_configurations_to_raw(
                ip_configurations=new_values.get("ip_configurations")
            )
            need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_network_interfaces_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    network_interface_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        network_interface_name: Azure Network Interface name.
        resource_id: Azure Policy Definition resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "subscription_id": subscription_id,
        "network_interface_name": network_interface_name,
        "location": resource["location"],
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    properties = resource.get("properties")
    if properties.get("ipConfigurations") is not None:
        existing_ip_configurations_payload = convert_raw_ip_configurations_to_present(
            ip_configurations=properties["ipConfigurations"]
        )
        resource_translated["ip_configurations"] = existing_ip_configurations_payload
    return resource_translated


def convert_present_to_raw_network_interfaces(
    hub,
    location: str,
    ip_configurations: List[Dict[str, Any]],
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        location: Resource location.
        ip_configurations: A list of IPConfigurations of the network interface.
           Each ip configuration supports fields  ((name(str), required),
           (subnet_id(str, optional), (private_ip_address_version(str, optional),
           (private_ip_address_allocation(str, optional), private_ip_address(str, optional)
           and primary (boolean, optional)).
        tags: Resource tags.
    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "location": location,
    }
    if tags is not None:
        payload["tags"] = tags
    if ip_configurations is not None:
        ip_configurations_payload = convert_present_ip_configurations_to_raw(
            ip_configurations
        )
        payload["properties"] = {}
        payload["properties"]["ipConfigurations"] = ip_configurations_payload

    return payload


def convert_raw_ip_configurations_to_present(ip_configurations: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        ip_configurations(List[Dict[str, Any]]): A list of IPConfigurations of the network interface.

    Returns:
         An ip_configurations List that contains the parameters that match respective present function's input format.
    """
    present_ip_configurations: List = []
    for ip_configuration in ip_configurations:
        existing_ip_configuration_payload = {
            "name": ip_configuration.get("name"),
            "private_ip_address_allocation": ip_configuration["properties"][
                "privateIPAllocationMethod"
            ],
        }
        if ip_configuration["properties"].get("subnet") is not None:
            existing_ip_configuration_payload["subnet_id"] = ip_configuration[
                "properties"
            ]["subnet"]["id"]
        if ip_configuration["properties"].get("privateIPAddressVersion") is not None:
            existing_ip_configuration_payload[
                "private_ip_address_version"
            ] = ip_configuration["properties"]["privateIPAddressVersion"]
        if ip_configuration["properties"].get("privateIPAddress") is not None:
            existing_ip_configuration_payload["private_ip_address"] = ip_configuration[
                "properties"
            ]["privateIPAddress"]
        if ip_configuration["properties"].get("primary") is not None:
            existing_ip_configuration_payload["primary"] = ip_configuration[
                "properties"
            ]["primary"]
        present_ip_configurations.append(existing_ip_configuration_payload)
    return present_ip_configurations


def convert_present_ip_configurations_to_raw(ip_configurations: List[Dict[str, Any]]):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        ip_configurations(List[Dict[str, Any]], required): A list of IPConfigurations of the network interface.
          Each ip configuration supports fields  ((name(str), required),
          (subnet_id(str, optional), (private_ip_address_version(str, optional),
          (private_ip_address_allocation(str, optional), private_ip_address(str, optional)
          and primary (boolean, optional)).

    Returns:
        ipConfigurations List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    raw_ip_configurations: List = []
    for ip_configuration in ip_configurations:
        payload = {
            "name": ip_configuration["name"],
            "properties": {},
        }
        if ip_configuration.get("private_ip_address_allocation") is not None:
            payload["properties"]["privateIPAllocationMethod"] = ip_configuration[
                "private_ip_address_allocation"
            ]
        if ip_configuration.get("subnet_id") is not None:
            payload["properties"]["subnet"] = {"id": ip_configuration["subnet_id"]}
        if ip_configuration.get("private_ip_address_version") is not None:
            payload["properties"]["privateIPAddressVersion"] = ip_configuration[
                "private_ip_address_version"
            ]
        if ip_configuration.get("private_ip_address") is not None:
            payload["properties"]["privateIPAddress"] = ip_configuration[
                "private_ip_address"
            ]
        if ip_configuration.get("primary") is not None:
            payload["properties"]["primary"] = ip_configuration["primary"]
        raw_ip_configurations.append(payload)
    return raw_ip_configurations


def diff_network_ip_configurations(
    new_values: List[Dict[str, any]], resource: List[Dict[str, any]], sorting_key: str
):
    """
    Compares network interface ip configurations to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        new_values: (List[Dict[str, any]]) Present value which will be given as input
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False
    """
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        resource, key=lambda x: x[sorting_key]
    )



"""
idem-azure can autogenerate states as needed using pop-create.
Once autogenerated states are tested, they should be moved from idem_azure/states/azure to idem_azure/states/azure
"""


def __init__(hub):
    # Provides the ctx argument to all execution modules
    # which will have profile info from the account module
    hub.exec.azure.ACCT = ["azure"]
    hub.exec.azure.URL = "https://management.azure.com"



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
import json
from collections import OrderedDict
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 20}}


async def present(
    hub,
    ctx,
    name: str,
    alias: str,
    billing_scope: str,
    display_name: str,
    workload: str,
    resource_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Create or update Subscription

    Args:
        name(str): The identifier for this state.
        alias(str): The alias name of the subscription to create or update. Can include alphanumeric,
          underscore, parentheses, hyphen, period (except at end), and Unicode characters that match
          the allowed characters.Regex pattern: ^[-\w\._\(\)]+$.
        billing_scope: billing scope associated with billing account id and enrollment account id to create subscription
        display_name: subscription display name
        workload: type of workload where subscription needs to be created. Can be Production/DevTest
        resource_id: resource unique id

    Returns:
        Dict

    Examples:

        .. code-block:: sls

        azure.subscription.subscriptions.present:
            - name: create_subscription_3
            - alias: dupSubTest
            - billing_scope: /providers/Microsoft.Billing/billingAccounts/{billingAccountID}/enrollmentAccounts/{enrollmentAccountId}}
            - display_name: Subscription Display Name
            - workload: Production
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if resource_id is None:
        resource_id = f"/providers/Microsoft.Subscription/aliases/{alias}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
        success_codes=[200],
    )
    hub.log.debug(f"Get Subscription details : {response_get}")

    if not response_get["result"]:
        if response_get["status"] == 404:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "alias": alias,
                        "display_name": display_name,
                        "subscription_id": None,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.subscription.subscriptions '{name}'"
                )
                return result
            else:
                # PUT operation to create subscription
                payload = hub.tool.azure.subscription.subscriptions.convert_present_to_raw_subscription(
                    billing_scope=billing_scope,
                    display_name=display_name,
                    workload=workload,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

            result[
                "new_state"
            ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                alias=alias,
                display_name=display_name,
                subscription_id=None,
                resource_id=resource_id,
            )
            result["comment"].append(
                f"Created azure.subscription.subscriptions '{name}'"
            )
            return result
        else:
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result
    else:
        # Update/rename subscription display name if given subscription already present
        # and display name is modified in the input
        subscription_id = response_get["ret"]["properties"]["subscriptionId"]
        response_get_detailed_subscription = await hub.exec.request.json.get(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}?api-version=2020-01-01",
            success_codes=[200],
        )
        if not response_get_detailed_subscription["result"]:
            hub.log.debug(
                f"Unable to fetch Azure subscription details {response_get_detailed_subscription['comment']}"
                f" {response_get_detailed_subscription['ret']}"
            )
            result["result"] = False
            result["comment"] = [
                response_get_detailed_subscription["comment"],
                response_get_detailed_subscription["ret"],
            ]
            return result
        existing_resource = response_get_detailed_subscription["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            alias=alias,
            display_name=None,
            subscription_id=None,
            resource_id=resource_id,
        )
        # Generate a new PUT operation payload with new values
        is_display_name_updated = display_name != existing_resource.get("displayName")
        if ctx.get("test", False):
            if not is_display_name_updated:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.subscription.subscriptions '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
                    resource=response_get["ret"],
                    idem_resource_name=name,
                    alias=alias,
                    display_name=display_name,
                    subscription_id=None,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Would update azure.subscription.subscriptions '{name}'"
                )
            return result
        # POST operation to rename subscription display_name
        if not is_display_name_updated:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.subscription.subscriptions '{name}' has no property need to be updated."
            )
            return result
        rename_payload = {"subscriptionName": display_name}
        response_rename = await hub.exec.request.json.post(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Subscription/rename?api-version=2020-09-01",
            success_codes=[200, 201],
            json=rename_payload,
        )
        if not response_rename["result"]:
            hub.log.debug(
                f"Could not update azure.subscription.subscriptions {response_rename['comment']} {response_rename['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_rename["comment"], response_rename["ret"]]
            return result

        result["new_state"] = copy.deepcopy(result["old_state"])
        result["new_state"]["display_name"] = display_name
        result["comment"].append(f"Updated azure.subscription.subscriptions '{name}'")
        return result


async def absent(hub, ctx, name: str, alias: str) -> Dict:
    r"""
    **Autogenerated function**

    Delete Subscription. This state disables the subscription and deletes respective alias.
    Once subscription is in disabled state, after 90 days it gets deleted automatically.

    Args:
        name(str): The identifier for this state.
        alias(str): The alias name of the subscription to delete.

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            subscription_is_absent:
              azure.subscription.subscriptions.absent:
                - name: value
                - alias: value
    """

    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    resource_id = f"/providers/Microsoft.Subscription/aliases/{alias}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
        success_codes=[200],
    )
    if response_get["result"]:
        subscription_id = response_get["ret"]["properties"]["subscriptionId"]
        response_get = await hub.exec.request.json.get(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}?api-version=2020-01-01",
            success_codes=[200],
        )
        result[
            "old_state"
        ] = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            alias=alias,
            display_name=None,
            subscription_id=subscription_id,
            resource_id=resource_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.subscription.subscriptions '{name}'"
            )
            return result

        # Cancel the subscription before deleting
        # Once subscription is cancelled, it will be in disabled state and after 90 days it gets deleted automatically.
        response_cancel = await hub.exec.request.json.post(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Subscription/cancel?api-version=2020-09-01",
            success_codes=[200, 202],
        )
        if not response_cancel["result"]:
            result["result"] = False
            result["comment"] = [response_cancel["comment"], response_cancel["ret"]]
            return result

        # Delete subscription alias.
        # After deleting the subscription with alias, deleted alias can be used to create new subscription
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-09-01",
            success_codes=[200, 202],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.subscription.subscriptions {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.subscription.subscriptions '{name}'")
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the management group is not associated with given subscription.
        result["comment"].append(
            f"azure.subscriptions.subscription '{name}' already absent"
        )
        return result
    else:
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe subscriptions in a way that can be recreated/managed with the corresponding "present" function

    Lists all subscriptions.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            $ idem describe azure.subscription.subscriptions
    """

    result = {}
    uri_parameters = OrderedDict({"subscriptions": "subscription_id"})
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions?api-version=2020-01-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value", None)
        hub.log.debug(f"Subscriptions : {json.dumps(resource_list, indent=1)}")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource["id"], uri_parameters
                )
                resource_translated = hub.tool.azure.subscription.subscriptions.convert_raw_subscription_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    alias=None,
                    display_name=None,
                    **uri_parameter_values,
                )
                result[resource_id] = {
                    f"azure.subscription.subscriptions.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    policy_definition_name: str,
    policy_type: str,
    mode: str,
    subscription_id: str = None,
    display_name: str = None,
    description: str = None,
    policy_rule: Dict = None,
    metadata: Dict = None,
    parameters: Dict = None,
    resource_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Create or update Policy Definitions

    Args:
        name(str): The identifier for this state.
        policy_definition_name(str): The name of the policy Definition.
        policy_type(str): The policy type. Possible values are BuiltIn, Custom and NotSpecified.
        mode(str): The policy mode that allows you to specify which resource types will be evaluated.
                   Some examples are All, Indexed, Microsoft.KeyVault.Data.
        subscription_id(str, optional): Subscription Unique id.
        display_name(str): The display name of the policy definition.
        description(str, optional): The description of the policy definition.
        policy_rule(dict, optional): The policy rule for the policy definition.
        metadata(dict, optional): The metadata for the policy definition.
        parameters(dict, optional): Parameters for the policy definition.
        resource_id(str, optional): Policy Definition resource id on Azure

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            policy_definition_is_present:
              azure.policy.policy_definitions.present:
                - name: value
                - policy_definition_name: value
                - subscription_id: value
                - policy_type: value
                - mode: value
                - display_name: value
                - description: value
                - metadata:
                        version: 1.0.0
                        category: RoleDefinitions
                - parameters:
                    roleDefinitionIds:
                        type: Array
                        metadata:
                            displayName: Approved Role Definitions
                            description: The list of role definition Ids.
                            strongType: roleDefinitionIds
                - policy_rule:
                        if:
                          allOf:
                            - field: type
                              equals: Microsoft.Authorization/roleAssignments
                            - not:
                                field: Microsoft.Authorization/roleAssignments/roleDefinitionId
                                in: "[parameters('roleDefinitionIds')]"
                        then:
                      effect: deny



    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/providers/Microsoft.Authorization/policyDefinitions/{policy_definition_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] == 404:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "policy_definition_name": policy_definition_name,
                        "subscription_id": subscription_id,
                        "policy_type": policy_type,
                        "mode": mode,
                        "display_name": display_name,
                        "description": description,
                        "policy_rule": policy_rule,
                        "metadata": metadata,
                        "parameters": parameters,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.policy.policy_definitions '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.policy.policy_definition.convert_present_to_raw_policy_definition(
                    policy_type=policy_type,
                    description=description,
                    display_name=display_name,
                    mode=mode,
                    metadata=metadata,
                    policy_rule=policy_rule,
                    parameters=parameters,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create Policy Definitions {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    policy_definition_name=policy_definition_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Created azure.policy.policy_definitions '{name}'"
                )
                return result

        else:
            hub.log.debug(
                f"Could not get azure.policy.policy_definitions {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result

    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            policy_definition_name=policy_definition_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        # Generate a new PUT operation payload with new values
        new_payload = (
            hub.tool.azure.policy.policy_definition.update_policy_definition_payload(
                existing_resource,
                {
                    "policy_type": policy_type,
                    "display_name": display_name,
                    "description": description,
                    "mode": mode,
                    "policy_rule": policy_rule,
                    "metadata": metadata,
                    "parameters": parameters,
                },
            )
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.policy.policy_definitions '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    policy_definition_name=policy_definition_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Would update azure.policy.policy_definitions '{name}'"
                )
            return result

        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.policy.policy_definitions '{name}' has no property need to be updated."
            )
            return result
        result["comment"] = result["comment"] + new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 201],
            json=new_payload["ret"],
        )
        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.policy.policy_definitions {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_put["comment"], response_put["ret"]]
            return result

        result[
            "new_state"
        ] = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            policy_definition_name=policy_definition_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        result["comment"].append(f"Updated azure.policy.policy_definitions '{name}'")
        return result


async def absent(
    hub, ctx, name: str, policy_definition_name: str, subscription_id: str = None
) -> Dict:
    r"""
    **Autogenerated function**

    Delete Policy Definition

    Args:
        name(str): The identifier for this state.
        policy_definition_name(str): The name of the policy definition to delete.
        subscription_id(str, optional): Subscription Unique id.

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.policy.policy_definitions.absent:
                - name: value
                - policy_definition_name: value
                - subscription_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/providers/Microsoft.Authorization/policyDefinitions/{policy_definition_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
        success_codes=[200],
    )
    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            policy_definition_name=policy_definition_name,
            resource_id=resource_id,
            subscription_id=subscription_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.policy.policy_definitions '{name}'"
            )
            return result
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 202],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.policy.policy_definitions {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.policy.policy_definitions '{name}'")
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.policy.policy_definitions '{name}' already absent"
        )
        return result
    else:
        hub.log.debug(
            f"Could not get azure.policy.policy_definitions '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Policy Definitions under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.policy.policy_definitions
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "policyDefinitions": "policy_definition_name",
            "subscriptions": "subscription_id",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Authorization/policyDefinitions?api-version=2021-06-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource["id"], uri_parameters
                )
                resource_translated = hub.tool.azure.policy.policy_definition.convert_raw_policy_definition_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    **uri_parameter_values,
                )
                result[resource_id] = {
                    f"azure.policy.policy_definitions.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    policy_assignment_name: str,
    policy_definition_id: str,
    parameters: Dict[str, Any] = None,
    resource_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Create or update Policy Assignments

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name which should be used for this Policy Assignment. Changing this forces a new Resource Policy Assignment to be created.
        policy_definition_id(str): The ID of the Policy Definition or Policy Definition Set. Changing this forces a new Policy Assignment to be created
        parameters(Dict, optional): Policy assignment parameters with respect to policy definition rule. Defaults to {}.
        resource_id(str, optional): Policy Assignment resource id on Azure

    Returns:
        dict

    Examples:

        .. code-block:: sls

            resource_is_present:
              azure.policy.policy_assignments.present:
                - name: value
                - scope: value
                - policy_assignment_name: value
                - policy_definition_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if parameters is None:
        parameters = {}

    if resource_id is None:
        resource_id = f"{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}"

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/{resource_id}?api-version=2021-06-01",
        success_codes=[200],
    )

    if not response_get["result"]:
        if response_get["status"] == 404:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "policy_assignment_name": policy_assignment_name,
                        "policy_definition_id": policy_definition_id,
                        "scope": scope,
                        "parameters": parameters,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.policy.policy_assignments '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.policy.policy_assignment.convert_present_to_raw_policy_assignment(
                    policy_definition_id=policy_definition_id,
                    parameters=parameters,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.policy.policy_assignments {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Created azure.policy.policy_assignments '{name}'"
                )
                return result

        else:
            hub.log.debug(
                f"Could not get azure.policy.policy_assignments {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result

    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            resource_id=resource_id,
        )
        # Generate a new PUT operation payload with new values
        new_payload = (
            hub.tool.azure.policy.policy_assignment.update_policy_assignment_payload(
                existing_resource,
                {
                    "policy_definition_id": policy_definition_id,
                    "parameters": parameters,
                },
            )
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.policy.policy_assignments '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Would update azure.policy.policy_assignments '{name}'"
                )
            return result

        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.policy.policy_assignments '{name}' has no property need to be updated."
            )
            return result
        result["comment"] = result["comment"] + new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 201],
            json=new_payload["ret"],
        )
        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.policy.policy_assignments {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_put["comment"], response_put["ret"]]
            return result

        result[
            "new_state"
        ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            resource_id=resource_id,
        )
        result["comment"].append(f"Updated azure.policy.policy_assignments '{name}'")
        return result


async def absent(hub, ctx, name: str, scope: str, policy_assignment_name: str) -> dict:
    r"""
    **Autogenerated function**

    Delete Policy Assignment

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name of the policy assignment to delete.

    Returns:
        dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.policy.policy_assignments.absent:
                - name: value
                - scope: value
                - policy_assignment_name: value
    """

    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    resource_id = f"{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}"
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/{resource_id}?api-version=2021-06-01",
        success_codes=[200],
    )
    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            resource_id=resource_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.policy.policy_assignments '{name}'"
            )
            return result
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
            success_codes=[200, 202],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.policy.policy_assignments {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(f"Deleted azure.policy.policy_assignments '{name}'")
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.policy.policy_assignments '{name}' already absent"
        )
        return result
    else:
        hub.log.debug(
            f"Could not get azure.policy.policy_assignments '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Policy Assignments under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.policy.policy_assignments
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Authorization/policyAssignments?api-version=2021-06-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                resource_translated = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                )
                result[resource_id] = {
                    f"azure.policy.policy_assignments.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 20}}


async def present(
    hub,
    ctx,
    name: str,
    management_group_name: str,
    resource_id: str = None,
    display_name: str = None,
    parent_id: str = None,
) -> Dict:
    r"""
    **Autogenerated function**

    Create or update Management Groups

    Args:
        name(str): The identifier for this state.
        management_group_name(str): The name of the management group to create or update. Can include alphanumeric, underscore, parentheses, hyphen, period (except at end), and Unicode characters that match the allowed characters.Regex pattern: ^[-\w\._\(\)]+$.
        display_name(str, Optional): The management group name to be displayed
        parent_id(str, Optional): creates management group under this id
        resource_id(str, Optional): Management group resource id on Azure

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            my-management-group:
              azure.management_groups.management_groups.present:
                - name: my-management-group
                - management_group_name: my-management-group-1
                - display_name: my-management-group-1
                - parent_id: /providers/Microsoft.Management/managementGroups/parent-management-group

    """

    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if resource_id is None:
        resource_id = (
            f"/providers/Microsoft.Management/managementGroups/{management_group_name}"
        )

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        # The response code is 404 or 403 means the management group does not exist or Authorization is failed
        if response_get["status"] in [404, 403]:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "management_group_name": management_group_name,
                        "resource_id": resource_id,
                        "display_name": display_name,
                        "parent_id": parent_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.management_groups.management_groups '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.management_groups.management_groups.convert_present_to_raw_management_group(
                    display_name=display_name, parent_id=parent_id
                )

                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
                    success_codes=[200, 201, 202],
                    json=payload,
                )

                if not response_put["result"]:
                    hub.log.error(
                        f"Could not create Management Group {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
                    resource=response_put,
                    idem_resource_name=name,
                    management_group_name=management_group_name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Created azure.management_groups.management_groups '{name}'"
                )
                return result

        else:
            hub.log.debug(
                f"Could not get azure.management_groups.management_groups {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_get["comment"], response_get["ret"]]
            return result

    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            management_group_name=management_group_name,
            resource_id=resource_id,
        )
        # Generate a new PUT operation payload with new values
        new_payload = hub.tool.azure.management_groups.management_groups.update_management_groups_payload(
            existing_resource, {"display_name": display_name, "parent_id": parent_id}
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.management_groups.management_groups '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    management_group_name=management_group_name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Would update azure.management_groups.management_groups '{name}'"
                )
            return result
        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.management_groups.management_groups '{name}' has no property need to be updated."
            )
            return result
        result["comment"] = result["comment"] + new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
            success_codes=[200],
            json=new_payload["ret"],
        )

        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.management_groups.management_groups {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_put["comment"], response_put["ret"]]
            return result

        result[
            "new_state"
        ] = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            management_group_name=management_group_name,
            resource_id=resource_id,
        )
        result["comment"].append(
            f"Updated azure.management_groups.management_groups '{name}'"
        )
        return result


async def absent(hub, ctx, name: str, management_group_name: str) -> Dict:
    r"""
    **Autogenerated function**

    Delete Management Groups

    Args:
        name(str): The identifier for this state.
        management_group_name(str): The name of the resource group to delete. The name is case insensitive.Regex pattern: ^[-\w\._\(\)]+$.

    Returns:
        Dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.management_groups.management_groups.absent:
                - name: value
                - management_group_name: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    resource_id = (
        f"/providers/Microsoft.Management/managementGroups/{management_group_name}"
    )
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
        success_codes=[200],
    )

    if response_get["result"]:
        result[
            "old_state"
        ] = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
            resource=response_get["ret"],
            idem_resource_name=name,
            management_group_name=management_group_name,
            resource_id=resource_id,
        )
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.management_groups.management_groups '{name}'"
            )
            return result

        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2020-05-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.management_groups.management_groups {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = [response_delete["comment"], response_delete["ret"]]
            return result

        result["comment"].append(
            f"Deleted azure.management_groups.management_groups '{name}'"
        )
        return result

    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.management_groups.management_groups '{name}' already absent"
        )
        return result
    else:
        hub.log.debug(
            f"Could not get azure.management_groups.management_groups '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Management Groups


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure.management_groups.management_groups
    """

    result = {}
    uri_parameters = OrderedDict({"managementGroups": "management_group_name"})
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/providers/Microsoft.Management/managementGroups?api-version=2020-05-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                resource_translated = hub.tool.azure.management_groups.management_groups.convert_raw_management_group_to_present(
                    resource=resource,
                    idem_resource_name=resource_id,
                    resource_id=resource_id,
                    **uri_parameter_values,
                )

                result[resource_id] = {
                    f"azure.management_groups.management_groups.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }
    return result
